# -*-python-*-
# $Id: s60-upcalendar.py,v 1.17 2006/08/22 18:16:46 asc Exp $

__package__    = "s60-upcalendar"
__version__    = "0.11"
__author__     = "Aaron Straup Cope"
__url__        = "http://www.aaronland.info/python/upcalendar/"
__cvsversion__ = "$Revision: 1.17 $"
__date__       = "$Date: 2006/08/22 18:16:46 $"
__copyright__  = "Copyright (c) 2006 Aaron Straup Cope. Perl Artistic License."

import e32
import appuifw
import calendar
from elementtree.ElementTree import XML, tostring, fromstring
from pdis.xpath import compile

import re
import anydbm
import os.path
import urllib
import httplib
import time

if not e32.in_emulator() :
    import sysinfo
    import calendar
    
#
#
#
    
class Upcalendar :

    def __init__ (self) :

        self.__upcoming = u"upcoming.org"
        self.__endpoint = u"/services/rest/"
        self.__auth     = u"/services/auth/"        
        self.__api_key   = u"b92238793f"

        self.__upcalendar = None
            
        self.__cfgpath = u"c:\\upcalendar.db"     
        self.lock = e32.Ao_lock()

        self.__username = u""
        self.msglog = u""
        
        self.old_title = appuifw.app.title
        appuifw.app.title = u"upcalendar"

        self.exit_flag = False
        appuifw.app.exit_key_handler = self.abort

        self.__options = [(u"Merge your watchlist", self.collect_events),
                          (u"Set default alarm", self.set_default_alarm),
                          (u"Change your auth token", self.get_auth_token),
                          (u"Purge metro and venue cache", self.purge_upcache)]
        
    #
    #
    #

    def loop (self) :

        if self.setup() :
            self.run()

        self.lock.wait()

    #
    #
    #
    
    def abort(self) :
        self.store_cfg()
        self.lock.signal()
        self.old_title = appuifw.app.title

    #
    #
    #
    
    def open_cfg (self) :
        self.log("Reading configs")

        if not os.path.exists(self.__cfgpath) :
            self.__cfg = anydbm.open(self.__cfgpath, "c")
            self.__cfg.close()

        self.__cfg = anydbm.open(self.__cfgpath, "w")
        return True

    #
    #
    #

    def purge_upcache (self, type='c|s|m|v', id ='\d+') :

        self.log("Purge upcoming metro and venue cache")
        
        pattern = '^up:(?:%s):%s$' % (type, id)
        
        for key, data in self.__cfg.items() :
            if re.match(pattern, key) :
                self.log("Remove %s" % key)
                del(self.__cfg[key])

        self.log("Purge complete")
        
    #
    #
    #

    def upcache_key (self, type, id) :
        return "up:%s:%s" % (type, id)

    #
    #
    #
    
    def upcache (self, type, id, data) :
        key = self.upcache_key(type, id)

        self.__cfg[key] = tostring(data)
        return True

    #
    #
    #
    
    def retrieve_upcache (self, type, id) :

        key = self.upcache_key(type, id)        

        if not self.__cfg.has_key(key) :
            return None

        try :
            return fromstring(self.__cfg[key])
        except Exception, e :
            del(self.__cfg[key])
            return None
        
    #
    #
    #
    
    def store_cfg (self) :
        self.__cfg.close()

    #
    #
    #

    def setup (self) :
        
        self.log("upcalendar (%s) says hello!" % __version__)
        self.log("---------------------")

        if not self.open_cfg() :
            return False

        if not self.open_calendar() :
            return False

        if not self.test_connection() :
            return False

        if not self.test_auth_token() :
            return False
        
        appuifw.app.menu = self.__options
        return True
    
    #
    #
    #
    
    def run (self) :
        self.log("----------------------------------")
        self.log("Welcome, %s - upcalendar is ready to go!" % self.__username)
        self.log("To begin, select 'Merge your watchlist' from the Options menu")
        
    #
    #
    #
    
    def collect_events (self) :

        self.log("Collecting events on your watchlist")
        
        xp_wl = compile("/rsp/watchlist")
        wl    = self.execute_upcoming_request("watchlist.getList")

        ids = []
        
        for event in xp_wl.evaluate(wl) :
            ids.append(event.attrib['event_id'])

        str_ids = u",".join(ids)

        ev = self.execute_upcoming_request("event.getInfo", event_id=str_ids)
        xp_ev = compile("/rsp/event")
        
        for event in xp_ev.evaluate(ev) :
            self.log("fetch event #%s" % event.attrib['id'])
            self.fetch_event(event)            
                
        self.log("Done!")
        
    #
    #
    #
    
    def fetch_event (self, event) :
                    
        vid   = event.attrib['venue_id']
        mid   = event.attrib['metro_id']
    
        #
        
        venue = self.retrieve_upcache('v', vid)
        
        if venue == None :
            vn    = self.execute_upcoming_request("venue.getInfo", venue_id=vid)
            xp_vn = compile("/rsp/venue")
            venue = xp_vn.evaluate(vn)

            if len(venue) > 0 :
                venue = venue[0]
                self.upcache('v', vid, venue)                
            else :
                venue = None
    
        #
            
        metro = self.retrieve_upcache('m', mid)
            
        if metro == None :
            mt    = self.execute_upcoming_request("metro.getInfo", metro_id=mid)
            xp_mt = compile("/rsp/metro")
            metro = xp_mt.evaluate(mt)

            if len(metro) > 0 :
                metro = metro[0]
                self.upcache('m', mid, metro)                
            else :
                metro = None
        #

        data = self.aggregate_event_data(event, venue, metro)
            
        if e32.in_emulator() :
            self.dump_event(data)
        else :
            self.store_event(data)
        
    #
    #
    #

    def open_calendar (self) :

        if e32.in_emulator () :
            return True
        
        try :
            cal = calendar.open()
        except Exception, e :
            self.log(e)
            return False

        self.__upcalendar = cal
        return True

    #
    #
    #

    def test_connection (self) :

        self.log("Testing network connection")
        
        if not e32.in_emulator() and sysinfo.signal() == 0 :
            self.error("No signal")
            return False

        head = self.execute_http_request(self.__upcoming, '/', 'HEAD')

        if head == False :
            self.error("Unable to connect to upcoming")
            return False
        
        return True
    
    #
    #
    #

    def open_url (self, url) :

        if e32.in_emulator() :
            self.log("open url : %s" % url)
            return True
        
        path   = 'z:\\System\\Apps\\Browser\\Browser.app'
        apprun = 'z:\\system\\programs\\apprun.exe'

        apprun_arg = path + ' "' + url + '"'
        e32.start_exe(apprun, apprun_arg, 1)

    #
    #
    #
    
    def get_auth_token (self) :

        options = [u"Enter auth frob by hand",
                   u"Get an auth token from scratch",
                   u"Huh? I don't even want to be here!"]

        prompt   = u"Configure your upcoming authentication"
        has_frob = appuifw.popup_menu(options, prompt)
    
        if has_frob == None :
            return self.get_auth_token()
        
        elif has_frob == 0 :
            
            prompt = u"Please enter your Upcoming auth frob"
            frob = appuifw.query(prompt, "text")
        
            if frob == None :
                self.log("New auth token request cancelled.")
                return None

            token = self.get_auth_token_for_frob(frob)
            
        elif has_frob == 1 :
            token = self.get_auth_token_www()

        else :
            return True
        
        if token == None :
            return False

        #
        
        self.__cfg['auth_token'] = unicode(token)

        self.log("Your authentication token has been updated!")
        return True
    
    #
    #
    #
    
    def get_auth_token_www (self) :

        appuifw.note(u"This might be painful depending on your phone.", "note")
        
        args = urllib.urlencode({'api_key':self.__api_key})
        url  = "http://%s%s?%s" % (self.__upcoming, self.__auth, args)
        
        prompt = u"You will need to visit the upcoming.org website. Launch browser?"

        if appuifw.query(prompt, "query") == None :
            self.log("New auth token request cancelled.")
            return None

        #
        
        self.open_url(url)

        
        prompt = u"Please enter your Upcoming auth frob"
        auth_frob = appuifw.query(prompt, "text")
        
        if auth_frob == None :
            self.log("New auth token request cancelled.")
            return None

        #

        return self.get_auth_token_for_frob(auth_frob)

    #
    #
    #
    
    def get_auth_token_for_frob (self, auth_frob) :
        
        xml = self.execute_upcoming_request('auth.getToken', frob=auth_frob)

        if xml == None :
            self.log("New auth token request unable to complete.")            
            return None
        
        xp = compile("/rsp/token")
        tok = xp.evaluate(xml)

        if tok == None :
            self.log("Hrm...upcoming sent back gibberish.")                        
            self.log("New auth token request unable to complete.")            
            return None

        #

        return tok[0].attrib['token']    

    #
    #
    #
    
    def test_auth_token (self) :

        self.log("Testing authentication token")
        
        if not self.__cfg.has_key('auth_token') :
            if not self.get_auth_token() :
                return False

        xml = self.execute_upcoming_request('auth.checkToken')

        if xml == None :
            self.log("Auth token failed validation")

            if appuifw.query("Re-enter auth token", query) :
                if self.get_auth_token() :
                    return self.test_auth_token()
                
            return False

        xp    = compile("/rsp/token")
        token = xp.evaluate(xml)[0]

        self.__username = token.attrib['user_username']
        return True

    #
    #
    #

    def aggregate_event_data (self, event, venue, metro) :

        aggr = {'location':'', 'start':0, 'end':0, 'description':'', 'uri':{}}

        if not venue == None :
            aggr['location'] = "%s, %s (%s)" % (venue.attrib['name'], venue.attrib['address'], venue.attrib['city'])

            if venue.attrib['phone'] != '' :
                aggr['location']  = "%s / %s" % (aggr['location'], venue.attrib['phone'])
                
        else :
            aggr['location'] = "(unknown location)"

        aggr['start_ymd']  = event.attrib['start_date']
        aggr['end_ymd']    = event.attrib['end_date']        
        aggr['start_time'] = event.attrib['start_time']
        aggr['end_time']   = event.attrib['end_time']        
        
        aggr['start'] = self.str_to_ts(event.attrib['start_date'], event.attrib['start_time'])
        aggr['end']   = aggr['start']

        if event.attrib['end_date'] != '' :
            aggr['end'] = self.str_to_ts(event.attrib['end_date'], event.attrib['end_time'])

            if aggr['end'] == 0 :
                aggr['end']   = aggr['start']

        if aggr['end'] < aggr['start'] :
            aggr['end']= aggr['start']
                
        aggr['description'] = event.attrib['name']

        aggr['uri']['upcoming'] = "http://upcoming.org/event/%s/" % event.attrib['id']
        return aggr
    
    #
    #
    #
    
    def dump_event (self, data) :

        self.log(data['description'])
        self.log(data['location'])
        self.log("%s - %s" % (data['start'], data['end']))

        if self.__cfg.has_key('alarm') and self.__cfg['alarm'] > 0 :
            alarm = data['start'] - int(self.__cfg['alarm'])
            self.log("alarm at %s" % alarm)
            
        self.log("\n")
        
    #
    #
    #

    def str_to_ts (self, str_date, str_time) :

        (y, m, d) = str_date.split('-')

        H = 12
        M = 0
        S = 0
        
        if str_time != '' :
            (H, M, S) = str_time.split(':')
        
        try :
            dt = time.mktime((int(y), int(m), int(d), int(H), int(M), int(S), 0, 0, 0))
            return int(dt)
        except Exception, e :
            self.log("bad date %s-%s-%sT%s:%s:%s (%s)" % (y, m, d, H, M, S, e))
            return 0
    
    #
    #
    #

    def find_event (self, data) :

        # self.log("find event %s, %s, %s" % (data['start'], data['end'], data['description']))

        try :
            found = self.__upcalendar.find_instances(data['start'], data['end'])
        except Exception, e :
            self.log("Local event search failed : %s" % e)
            return None
        
        if len(found) == 0 :
            return None

        for result in found :
            event = self.__upcalendar[result['id']]

            if event.content == data['description'] :
                return event
            
        return None
        
    #
    #
    #
    
    def store_event (self, data) :

        event = self.find_event(data)
    
        if event == None :
            event = self.__upcalendar.add_appointment()
        else :
            event.begin()
            
        try :
            self.update_event(event, data)
        except Exception, e :
            self.log("Failed to update event : %s" % e)
            return False

        try :
            event.commit()
        except Exception, e:
            self.log("Failed to commit event : %s" % e)
            return False

        self.log("Stored %s (%s)" % (event.content, data['start_ymd']))
        return True

    #
    #
    #

    def update_event (self, event, data) :

        event.set_time(data['start'], data['end'])
        event.content  = data['description']
        event.location = data['location']

        if self.__cfg.has_key('alarm') and self.__cfg['alarm'] > 0 :
            event.alarm = data['start'] - int(self.__cfg['alarm'])
            
        return True
    
    #
    #
    #

    def set_default_alarm (self) :

        input = appuifw.query(u"Default alarm for events", "text")
        alarm = 0

        if input == '0' :
            self.log("Alarm for events disabled. You will need to re-merge your watchlist for these changes to take effect.")
        elif input == None :
            pass
        else :
            input = input.lower()
            match = re.match("^(\d+)(m|h|d)$", input)

            if match == None :
                self.error("Invalid alarm. Alarms must be in the form of 'number' + 'm', 'h' or 'd' (minutes, hours, day)")
            else :
                (num, type) = match.groups()

                min  = 60
                hour = min * 60
                day  = hour * 24

                map     = {'m':min, 'h':hour, 'd':day}
                str_map = {'m':'minute', 'h':'hour', 'd':'day'}

                alarm = int(num) * int(map[type])
                
                interval = str_map[type]

                if num > 1 :
                    interval += "s"
                    
                self.log("Default alarm set to %s %s before each event. You will need to re-merge your watchlist for these changes to take effect." % (num, interval))
        
        self.__cfg['alarm'] = unicode(alarm)
        return True

    #
    #
    #
    
    def execute_upcoming_request (self, method, **kwargs) :

        kwargs['method']  = method
        kwargs['api_key'] = self.__api_key
        kwargs['token']   = self.__cfg['auth_token']
        
        req     = "%s?%s" % (self.__endpoint, urllib.urlencode(kwargs))
        content = self.execute_http_request(self.__upcoming, req)

        # self.log(content)
        return XML(content)

    #
    #
    #

    def execute_http_request (self, host, req, action='GET') :

        try :
            conn = httplib.HTTPConnection(host)
            conn.request(action, req)
        except Exception, e :
            self.log(e)
            return False

        res = conn.getresponse()
        
        if res.status != 200 :
            self.log(res.status)
            return False

        return res.read()
    
    #
    #
    #

    def error (self, msg) :
        self.log("[error] %s" % msg)

    #
    #
    #
    
    def log (self, msg) :

        self.msglog += "%s\n" % msg
        
        t = appuifw.Text()
        t.write(unicode(self.msglog))

        appuifw.app.body = t
        
    #
    #
    #
    
if __name__ == "__main__" :
    app = Upcalendar()
    app.loop()
